/*
mpcx_tree

high-level functions working with trees of polynomials

Copyright (C) 2009, 2010, 2011 Andreas Enge

This file is part of the MPFRCX Library.

The MPFRCX Library is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or (at your
option) any later version.

The MPFRCX Library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

You should have received a copy of the GNU Lesser General Public License
along with the MPC Library; see the file COPYING.LIB.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
MA 02111-1307, USA.
*/

#include "mpfrcx-impl.h"

static void mpcx_evaltree (mpcx_tree_ptr t, mpcx_tree_srcptr subproducts,
   mpcx_srcptr f);
static void mpcx_multieval_fast (mpc_t *values, mpc_t *args, int no,
   mpcx_srcptr f);

/**************************************************************************/

void mpcx_tree_init (mpcx_tree_ptr t, int no, mpfr_prec_t prec) {
   /* initialises a binary tree with "no" leaves for storing polynomials  */                        /* of precision "prec"                                                 */
   /* All tree nodes are allocated and initialised.                       */

   int i, j, tmp;

   t->no = no;
   for (i = no, t->levels = 1; i > 1; i = (i+1) / 2, t->levels++);
   t->node = (mpcx_t **) malloc (t->levels * sizeof (mpcx_t *));
   t->width = (int *) malloc (t->levels * sizeof (int));

   for (i = 0, tmp = no; i < t->levels; i++, tmp = (tmp + 1) / 2) {
      t->width [i] = tmp;
      t->node [i] = (mpcx_t *) malloc (tmp * sizeof (mpcx_t));
      for (j = 0; j < tmp; j++)
         mpcx_init (t->node [i][j], 2, prec);
   }
}

/**************************************************************************/

void mpcx_tree_clear (mpcx_tree_ptr t) {
   /* clears t and all the polynomials on its nodes                       */

   int i, j;

   for (i = 0; i < t->levels; i++) {
      for (j = 0; j < t->width [i]; j++)
         mpcx_clear (t->node [i][j]);
      free (t->node [i]);
   }
   free (t->node);
   free (t->width);
}

/**************************************************************************/

void mpcx_tree_get_root (mpcx_ptr f, mpcx_tree_srcptr t) {
   /* assigns the root of t to f */

   mpcx_set (f, t->node [t->levels-1][0]);
}

/**************************************************************************/

void mpcx_subproducttree (mpcx_tree_ptr t, mpcx_t *factors) {
   /* takes an array of t->no polynomials and returns via t the tree      */
   /* containing their subproducts: res [0,j] contains factors [j], and   */
   /* res [i,j] contains res [i-1,2j] * res [i-1,2j+1].                   */
   /* Optimal (at least if all degrees are the same) version as found out */
   /* with Guillaume and Pierrick: always multiply two factors, sweeping  */
   /* the array from left to right                                        */
   /* t must be initialised with the correct number of leaves (as many as */
   /* there are entries in "factors") and precision (usually the common   */
   /* precision of all entries in "factors").                             */

   int i, j;

   /* construct the basic layer */
   for (j = 0; j < t->width [0]; j++)
      mpcx_set (t->node [0][j], factors [j]);

   for (i = 1; i < t->levels; i++) {
      /* construct layer i */
      for (j = 0; j < t->width [i]; j++)
         if (j == t->width [i] - 1 && t->width [i-1] % 2 == 1)
            mpcx_set (t->node [i][j], t->node [i-1][2*j]);
         else
            mpcx_mul (t->node [i][j],
                      t->node [i-1][2*j], t->node [i-1][2*j+1]);
   }
}

/**************************************************************************/

static void mpcx_evaltree (mpcx_tree_ptr t, mpcx_tree_srcptr subproducts,
   mpcx_srcptr f) {
   /* second step in multipoint evaluation; takes a subproduct tree with  */
   /* "no" entries and a polynomial f, and returns in t the tree of the   */
   /* f mod subproducts [i][j], computed at the precision of f            */
   /* t must be initialised with the same number of leaves as the         */
   /* subproduct tree.                                                    */

   int      i, j;

   mpcx_rem (t->node [subproducts->levels-1][0],
             f, subproducts->node [t->levels-1][0]);
   for (i = t->levels - 2; i >= 0; i--)
      for (j = 0; j < t->width [i]; j++)
         mpcx_rem (t->node [i][j],
                   t->node [i+1][j/2], subproducts->node [i][j]);
}

/***************************************************************************/

static void mpcx_multieval_fast (mpc_t *values, mpc_t *args, int no,
                                 mpcx_srcptr f) {
   /* The internal version of the function. Always builds a subproduct    */
   /* tree independently of the degree of f and the number of arguments.  */

   mpcx_tree_t subproducts, evaltree;
   mpcx_t*  factors;
   int      i, j;

   factors = (mpcx_t *) malloc (no * sizeof (mpcx_t));
   for (i = 0; i < no; i++) {
      mpcx_init (factors [i], 2, f->prec);
      factors [i]->deg = 1;
      mpc_set_ui (factors [i]->coeff [1], 1, MPC_RNDNN);
      mpc_neg (factors [i]->coeff [0], args [i], MPC_RNDNN);
   }

   mpcx_tree_init (subproducts, no, factors [0]->prec);
   mpcx_subproducttree (subproducts, factors);

   mpcx_tree_init (evaltree, no, factors [0]->prec);
   mpcx_evaltree (evaltree, subproducts, f);

   for (j = 0; j < no; j++)
      mpc_set (values [j], evaltree->node [0][j]->coeff [0], MPC_RNDNN);

   mpcx_tree_clear (subproducts);
   mpcx_tree_clear (evaltree);

   for (i = 0; i < no; i++)
      mpcx_clear (factors [i]);
   free (factors);
}

/**************************************************************************/

void mpcx_multieval (mpc_t *values, mpc_t *args, int no, mpcx_srcptr f) {
   /* evaluates the polynomial f in the "no" arguments args and returns   */
   /* the results via values. values must contain sufficient space, and   */
   /* each entry must be initialised.                                     */
   /* values and args may point to the same array.                        */
   /* Works only if f and values have the same precision.                 */
   /* Splits the arguments into portions of about half the degree of f    */
   /* and calls the fast multipoint evaluation routine on the chunks.     */

   int chunks, size;
   int i;

   chunks = (2 * no) / f->deg;
   if (chunks == 0)
      chunks = 1;
   size = no / chunks;

   for (i = 0; i < chunks - 1; i++)
      mpcx_multieval_fast (values + i*size, args + i*size, size, f);
   mpcx_multieval_fast (values + (chunks - 1)*size, args + (chunks - 1)*size,
         no - (chunks - 1) * size, f);
}

/***************************************************************************/

void mpcx_hecke (mpcx_ptr rop, mpcx_tree_srcptr subproducts, mpcx_t *vals) {
   /* Let factors [0],... be the leaves of subproducts, no their number,   */
   /* and f=\prod_{i=0}^{no-1} factors [i] the polynomial at the root of   */
   /* subproducts.                                                         */
   /* Returns \sum_{i=0}^{no-1} val [i]*f/factors [i], which is a          */
   /* simplified case of Lagrange interpolation and occurs when computing  */
   /* the Hecke representation of an algebraic number with conjugates      */
   /* val [i] in a number field defined by f.                              */

   mpcx_tree_t hecke;
   mpcx_t tmp;
   mpfr_prec_t prec;
   int i, j;

   prec = subproducts->node [0][0]->prec;
   mpcx_tree_init (hecke, subproducts->no, prec);
   mpcx_init (tmp, subproducts->node [subproducts->levels-1][0]->deg, prec);
      /* rough estimate of the size, works for Hecke representation */

   for (j = 0; j < subproducts->no; j++)
      mpcx_set (hecke->node [0][j], vals [j]);
   for (i = 1; i < subproducts->levels; i++)
      for (j = 0; j < subproducts->width [i]; j++) {
         if (j == hecke->width [i] - 1 && hecke->width [i-1] % 2 == 1)
            mpcx_set (hecke->node [i][j], hecke->node [i-1][2*j]);
         else {
            mpcx_mul (hecke->node [i][j],
                     hecke->node [i-1][2*j], subproducts->node [i-1][2*j+1]);
            mpcx_mul (tmp,
                     hecke->node [i-1][2*j+1], subproducts->node [i-1][2*j]);
            mpcx_add (hecke->node [i][j], hecke->node [i][j], tmp);
         }
   }
   mpcx_set (rop, hecke->node [hecke->levels-1][0]);

   mpcx_tree_clear (hecke);
   mpcx_clear (tmp);
}

/***************************************************************************/
